import pickle
import re
import sys
import argparse
import torch
import json
import numpy
import scipy.sparse
import gzip
from torch.utils.data import DataLoader, Dataset
import functools
import numpy
import random
import logging
from torch.optim import Adam, SGD
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
import torch.nn.functional as F
from starcoder.dataset import NumericField, DistributionField, CategoricalField, SequentialField, IntegerField, DateField, Missing, Unknown


# (batch_count x (entity_representation_size + (bottleneck_size * relation_count)) :: Float) -> (batch_count x entity_representation_size :: Float)
class Autoencoder(torch.nn.Module):
    def __init__(self, sizes, activation):
        super(Autoencoder, self).__init__()
        self._encoding_layers = []
        self._decoding_layers = []
        for i in range(len(sizes) - 1):
            f = sizes[i] if isinstance(sizes[i], int) else sum(sizes[i])
            t = sizes[i + 1] if isinstance(sizes[i + 1], int) else sum(sizes[i + 1])
            a = sizes[i] if isinstance(sizes[i], int) else sizes[i][0]
            self._encoding_layers.append(torch.nn.Linear(f, t))
            self._decoding_layers.append(torch.nn.Linear(t, a))
        self._encoding_layers = [torch.nn.Identity()] if len(self._encoding_layers) == 0 else self._encoding_layers
        self._decoding_layers = [torch.nn.Identity()] if len(self._decoding_layers) == 0 else self._decoding_layers
        self._encoding_layers = torch.nn.ModuleList(self._encoding_layers)
        self._decoding_layers = torch.nn.ModuleList(reversed(self._decoding_layers))
        self._loss = torch.nn.MSELoss()
        self._activation = activation()
        #self._dropout = torch.nn.Dropout(p=0.5)
    def forward(self, x):
        y = x.clone()
        #x = self._dropout(x)
        for layer in self._encoding_layers:
            x = self._activation(layer(x)) #torch.nn.functional.leaky_relu(layer(x))
        bottleneck = x.clone().detach()
        for layer in self._decoding_layers:
            x = self._activation(layer(x)) #torch.nn.functional.leaky_relu(layer(x))
        return (x, bottleneck, self._loss(x, y))
    @property
    def input_size(self):
        return self._encoding_layers[0].in_features    
    @property
    def output_size(self):
        return self._decoding_layers[-1].out_features        


#
class Projector(torch.nn.Module):

    def __init__(self, input_size, output_size, activation):
        super(Projector, self).__init__()
        self._layer = torch.nn.Linear(input_size, output_size)

    def forward(self, x):
        return torch.nn.functional.leaky_relu(self._layer(x))


# (batch_size :: Int) -> (batch_size x field_representation_size :: Float)
class CategoricalEncoder(torch.nn.Module):
    def __init__(self, field, activation, **args):
        super(CategoricalEncoder, self).__init__()
        self._embeddings = torch.nn.Embedding(num_embeddings=len(field), embedding_dim=args.get("embedding_size", 32))

    def forward(self, x):
        #print(x.shape)
        #print(self._embeddings.num_embeddings)
        retval = self._embeddings(x) #.long())
        #print(retval)
        return(retval)

    @property
    def input_size(self):
        return 1

    @property
    def output_size(self):
        return self._embeddings.embedding_dim


# (batch_size x entity_representation_size :: Float) -> (batch_size x item_types :: Float)
class CategoricalDecoder(torch.nn.Module):
    def __init__(self, field, input_size, activation, **args): #input_size):
        super(CategoricalDecoder, self).__init__()
        output_size = len(field)
        self._layerA = torch.nn.Linear(input_size, output_size)
        self._layerB = torch.nn.Linear(output_size, output_size)
    def forward(self, x):
        x = torch.nn.functional.leaky_relu(self._layerA(x))
        x = self._layerB(x)
        x = torch.nn.functional.log_softmax(x, dim=1)
        return x
    @property
    def input_size(self):
        return self._layer.in_features
    @property
    def output_size(self):
        return self._layer.out_features        


CategoricalLoss = torch.nn.NLLLoss


# (batch_count :: Float) -> (batch_count :: Float)
class NumericEncoder(torch.nn.Module):
    def __init__(self, field, activation, **args):
        super(NumericEncoder, self).__init__()
    def forward(self, x):
        retval = torch.as_tensor(torch.unsqueeze(x, 1), dtype=torch.float32, device=x.device)
        return retval
    @property
    def input_size(self):
        return 1
    @property
    def output_size(self):
        return 1


# (batch_count x entity_representation_size :: Float) -> (batch_count :: Float)    
class NumericDecoder(torch.nn.Module):
    def __init__(self, field, input_size, activation, **args):
        super(NumericDecoder, self).__init__()
        self._linear = torch.nn.Linear(input_size, 1)
    def forward(self, x):
        retval = torch.nn.functional.leaky_relu(self._linear(x))
        retval = self._linear(torch.nn.functional.leaky_relu(x))
        return retval.squeeze()
    @property
    def input_size(self):
        return self._linear.in_features
    @property
    def output_size(self):
        return self._linear.out_features


NumericLoss = torch.nn.MSELoss


# (batch_count :: Float) -> (batch_count :: Float)
class DistributionEncoder(torch.nn.Module):
    def __init__(self, field, activation, **args):
        self._size = len(field._categories)
        super(DistributionEncoder, self).__init__()
    def forward(self, x):
        return x #torch.unsqueeze(x, 1)
    @property
    def input_size(self):
        return self._size
    @property
    def output_size(self):
        return self._size


# (batch_count x entity_representation_size :: Float) -> (batch_count :: Float)    
class DistributionDecoder(torch.nn.Module):
    def __init__(self, field, input_size, activation, **args):
        super(DistributionDecoder, self).__init__()
        self._linear = torch.nn.Linear(input_size, len(field._categories))
    def forward(self, x):
        return torch.nn.functional.log_softmax(torch.nn.functional.leaky_relu(self._linear(x)).squeeze(), dim=1)
    @property
    def input_size(self):
        return self._linear.in_features
    @property
    def output_size(self):
        return self._linear.out_features


DistributionLoss = torch.nn.KLDivLoss


# item_sequences -> lengths -> hidden_state
# (batch_count x max_length :: Int) -> (batch_count :: Int) -> (batch_count x entity_representation_size :: Float)
class SequentialEncoder(torch.nn.Module):
    def __init__(self, field, activation, **args):
        super(SequentialEncoder, self).__init__()
        es = args.get("embedding_size", 32)
        hs = args.get("hidden_size", 64)
        rnn_type = args.get("rnn_type", torch.nn.GRU)
        self._embeddings = torch.nn.Embedding(num_embeddings=len(field), embedding_dim=es)
        self._rnn = rnn_type(es, hs, batch_first=True, bidirectional=False)
    def forward(self, x):
        l = (x != Missing.value).sum(1)
        nonempty = l != 0
        x = x[nonempty]
        l = l[nonempty]
        embs = self._embeddings(x)
        pk = torch.nn.utils.rnn.pack_padded_sequence(embs, l, batch_first=True, enforce_sorted=False)
        output, h = self._rnn(pk)
        h = h.squeeze(0)
        retval = torch.zeros(size=(nonempty.shape[0], h.shape[1]), device=l.device)
        mask = nonempty.unsqueeze(1).expand((nonempty.shape[0], h.shape[1]))
        retval.masked_scatter_(mask, h)
        return retval
    @property
    def input_size(self):
        return self._rnn.input_size
    @property
    def output_size(self):
        return self._rnn.hidden_size


# representations -> item_distributions
# (batch_count x entity_representation_size :: Float) -> (batch_count x max_length x item_types :: Float)
class SequentialDecoder(torch.nn.Module):
    def __init__(self, field, input_size, activation, **args): #hidden_size, rnn_type=torch.nn.GRU):
        super(SequentialDecoder, self).__init__()
        hs = args.get("hidden_size", 128)
        rnn_type = args.get("rnn_type", torch.nn.GRU)
        self._max_length = field._max_length
        self._rnn = rnn_type(input_size, hs, batch_first=True, bidirectional=False)
        self._classifier = torch.nn.Linear(hs, len(field))
    def forward(self, x):
        x = torch.stack([x for i in range(self._max_length)], dim=1)
        output, _ = self._rnn(x)
        outs = []
        for i in range(self._max_length):
            outs.append(torch.nn.functional.log_softmax(self._classifier(output[:, i, :]), dim=1))
        return torch.stack(outs, 1)
    @property
    def input_size(self):
        return self._rnn.input_size
    @property
    def output_size(self):
        return self._rnn.hidden_size


class SequentialLoss(object):
    def __init__(self, reduction):
        self._nll = torch.nn.NLLLoss(reduction=reduction)
    def __call__(self, x, target):
        x = x[:, 0:target.shape[1], :]
        losses = []
        for v in range(target.shape[1]):
            losses.append(self._nll(x[:, v, :], target[:, v]))
        return torch.cat(losses)


# representations -> summary
# (related_entity_count x bottleneck_size) -> (bottleneck_size)
class RNNSummarizer(torch.nn.Module):
    def __init__(self, input_size, activation, rnn_type=torch.nn.GRU):
        super(RNNSummarizer, self).__init__()
        self._rnn = rnn_type(input_size, input_size, batch_first=True)
    def forward(self, representations):
        out, h = self._rnn(representations.unsqueeze(0))
        return h.squeeze()

class MaxPoolSummarizer(torch.nn.MaxPool1d):
    def __init__(self, input_size, activation):
        super(MaxPoolSummarizer, self).__init__(1)
    def forward(self, x):
        return super(MaxPoolSummarizer, self).forward(x)

class SingleSummarizer(torch.nn.Identity):
    def __init__(self, input_size, activation):
        self._input_size = input_size
        super(SingleSummarizer, self).__init__()
    def forward(self, x):
        if x.shape[0] == 0:
            return torch.zeros(shape=(self._input_size,))
        else:
            return x[0]
        

    
#class DummyAutoencoder(torch.nn.Identity):
#    def __init__(self):
#        super(DummyAutoencoder, self).__init__()
#    def forward(self, x):
#        return (x, None, None)


#class DummyProjector(torch.nn.Identity):
#    def __init__(self, size):
#        super(DummyProjector, self).__init__()
#        self._size = size
#    def forward(self, x):
#        return torch.zeros(size=(x.shape[0], self._size), device=x.device)

#class IdentityProjector(torch.nn.Identity):
#    pass

class MLPProjector(torch.nn.Module):
    def __init__(self, in_size, out_size, activation):
        super(MLPProjector, self).__init__()
        self._layer = torch.nn.Linear(in_size, out_size)
    def forward(self, x):
        return self._layer(x)

field_models = {
    NumericField : (NumericEncoder, NumericDecoder, NumericLoss(reduction="none")),
    DistributionField : (DistributionEncoder, DistributionDecoder, DistributionLoss(reduction="none")),
    IntegerField : (NumericEncoder, NumericDecoder, NumericLoss(reduction="none")),
    CategoricalField : (CategoricalEncoder, CategoricalDecoder, CategoricalLoss(reduction="none")),
    SequentialField : (SequentialEncoder, SequentialDecoder, SequentialLoss(reduction="none")),
    DateField : (NumericEncoder, NumericDecoder, NumericLoss(reduction="none")),
}


class GraphAutoencoder(torch.nn.Module):
    def __init__(self,
                 spec,
                 depth,
                 autoencoder_shapes,
                 embedding_size,
                 hidden_size,
                 field_dropout,
                 hidden_dropout,
                 summarizers=SingleSummarizer,
                 activation=torch.nn.ReLU,
                 projected_size=None):
        """
        """
        super(GraphAutoencoder, self).__init__()

        self._track_bottlenecks = {}
        
        self._device = torch.device('cpu')
        self._autoencoder_shapes = autoencoder_shapes

        self._field_dropout = field_dropout
        self._hidden_dropout = hidden_dropout
        self._embedding_size = embedding_size
        self._hidden_size = hidden_size
        self._bottleneck_size = None if autoencoder_shapes in [[], None] else autoencoder_shapes[-1]

        self._cache = {}
        
        # entity and field specification
        self._spec = spec

        # order in which to process entity-types
        self._entity_type_order = sorted(spec.entity_types)

        # order in which to process each entity-type's fields
        self._entity_type_field_order = {et : sorted([f for f in spec.entity_fields(et) if type(spec.field_object(f)) in field_models], key=lambda x : str(x)) for et in spec.entity_types}

        # list of non-ID/entity-type fields
        self._field_names = spec.data_field_names
        
        # order in which to process each entity-type's relations
        self._entity_type_relation_order = {et : sorted([x for x in spec.entity_relations(et)]) for et in spec.entity_types}
        
        # how many hops to propagate signal in the graph
        self._depth = depth        

        # encoded widths of entities
        self._boundary_sizes = {}

        # an encoder for each field
        self._field_encoders = {}
        for field_name in self._spec.field_names:
            field_type = type(self._spec.field_object(field_name))
            if field_type in field_models:
                self._field_encoders[field_name] = field_models[field_type][0](self._spec._field_name_to_object[field_name], activation)
        self._field_encoders = torch.nn.ModuleDict(self._field_encoders)

        for entity_type in self._entity_type_order:
            for field_name in self._entity_type_field_order[entity_type]:
                self._boundary_sizes[entity_type] = self._boundary_sizes.get(entity_type, 0)
                self._boundary_sizes[entity_type] += self._field_encoders[field_name].output_size

        #an autoencoder for each entity type and depth
        self._entity_autoencoders = {}        
        for entity_type in self._entity_type_order:
            boundary_size = self._boundary_sizes.get(entity_type, 0)
            self._entity_autoencoders[entity_type] = [Autoencoder([boundary_size] + self._autoencoder_shapes, activation)]
            for _ in self._spec.entity_relations(entity_type):
                boundary_size += self._bottleneck_size
            for depth in range(self._depth):
                self._entity_autoencoders[entity_type].append(Autoencoder([boundary_size] + self._autoencoder_shapes, activation))

            self._entity_autoencoders[entity_type] = torch.nn.ModuleList(self._entity_autoencoders[entity_type])
        self._entity_autoencoders = torch.nn.ModuleDict(self._entity_autoencoders)

        # a summarizer for each entity-to-entity relation
        if self._depth > 0:
            self._summarizers = {}
            for entity_type in self._entity_type_order:
                self._summarizers[entity_type] = {}
                for name in spec.entity_relations(entity_type):
                    self._summarizers[entity_type][name] = summarizers(self._bottleneck_size, activation)
                self._summarizers[entity_type] = torch.nn.ModuleDict(self._summarizers[entity_type])
            self._summarizers = torch.nn.ModuleDict(self._summarizers)

        # MLP for each entity type to project representations to a common size
        # note the largest boundary size
        self._projected_size = projected_size if projected_size != None else max(self._boundary_sizes.values())
        self._projectors = {}
        for entity_type in self._entity_type_order:
            boundary_size = self._boundary_sizes.get(entity_type, 0)
            for _ in self._spec.entity_relations(entity_type):
                boundary_size += self._bottleneck_size
            self._projectors[entity_type] = MLPProjector(boundary_size, self._projected_size, activation)
        self._projectors = torch.nn.ModuleDict(self._projectors)
        
        # a decoder for each field
        # change from per-entity-per-field to just per-field!
        self._field_decoders = {}
        for field_name in self._spec.field_names:
            field_type = type(self._spec.field_object(field_name))
            if field_type in field_models:
                self._field_decoders[field_name] = field_models[field_type][1](self._spec._field_name_to_object[field_name], self._projected_size, activation)
        self._field_decoders = torch.nn.ModuleDict(self._field_decoders)

    def cuda(self):
        self._device = torch.device('cuda:0')
        super(GraphAutoencoder, self).cuda()

    @property
    def entity_type_order(self):
        return self._entity_type_order
    
    @property
    def spec(self):
        return self._spec
        
    @property
    def id_field(self):
        return self.spec.id_field
    
    @property
    def device(self):
        return self._device

    @property
    def entity_type_field(self):
        return self.spec.entity_type_field

    @property
    def device(self):
        return self._device

    @property
    def entity_types(self):
        return self.spec.entity_types

    def encoder(self, field_name):
        return self._field_encoders[field_name]
    
    def forward(self, entities, adjacencies):
        logging.debug("Starting forward pass")

        num_entities = len(entities[self.id_field])
        entity_indices = {}
        field_indices = {}
        entity_field_indices = {}
        autoencoder_boundary_pairs = []

        logging.debug("Assembling entity, field, and (entity, field) indices")        
        index_space = torch.arange(0, entities[self.entity_type_field].shape[0], 1)
        for entity_type in self._entity_type_order:
            entity_mask = (entities[self.entity_type_field] == self._spec.field_object(self.entity_type_field)[entity_type])
            entity_indices[entity_type] = index_space.masked_select(entity_mask)
            for field_name in self._entity_type_field_order[entity_type]:
                field_mask = ~torch.isnan(torch.reshape(entities[field_name], (entities[field_name].shape[0], -1)).sum(1))
                field_indices[field_name] = index_space.masked_select(field_mask)
                entity_field_mask = entity_mask & field_mask
                entity_field_indices[(entity_type, field_name)] = index_space.masked_select(entity_field_mask)

        
        
        logging.debug("Encoding each input field to a fixed-length representation")
        field_encodings = {}
        for field_name in self.spec.data_field_names:
            field_encodings[field_name] = torch.full(size=(num_entities, self.encoder(field_name).output_size),
                                                     fill_value=0.0,
                                                     device=self.device,
                                                     dtype=torch.float32)
            indices = field_indices[field_name]
            field_values = torch.index_select(entities[field_name], 0, indices)
            field_encodings[field_name][indices] = self.encoder(field_name)(field_values)

        logging.debug("Constructing entity-autoencoder inputs by concatenating field encodings")
        autoencoder_inputs = {}
        for entity_type, indices in entity_indices.items():
            #
            # each appended value should have shape (entity_count x encoding_width)
            #
            autoencoder_inputs[entity_type] = []
            for field_name in self._entity_type_field_order[entity_type]:
                #ef_key = (entity_type, field_name)                
                # should missing values be 0, NaN...?
                #fill_value = 0 # float("nan") # self._spec.field_object(field_name).missing_value
                #encodings = torch.full(size=(num_entities, self._field_encoders[field_name].output_size),
                #                       fill_value=fill_value,
                #                       device=self._device, dtype=torch.float32)
                #encodings = torch.full(size=(entity_index.shape[0], self._field_encoders[field_name].output_size),
                #                       fill_value=fill_value,
                #                       device=self._device, dtype=torch.float32)

                #field_values = torch.index_select(entities[field_name], 0, entity_field_indices[ef_key])
                #field_values = torch.index_select(entities[field_name], 0, entity_indices[entity_type])
                #print(field_name, field_values, entity_field_indices)
                #encodings[entity_field_indices[ef_key]] = self._field_encoders[field_name](field_values)
                #encodings[entity_field_indices[ef_key]] = self._field_encoders[field_name](field_values)
                autoencoder_inputs[entity_type].append(torch.index_select(field_encodings[field_name], 0, indices))
                #autoencoder_inputs[entity_type].append(torch.index_select(encodings)
            autoencoder_inputs[entity_type].append(torch.zeros(size=(indices.shape[0], 0), dtype=torch.float32, device=self._device))
            autoencoder_inputs[entity_type] = torch.cat(autoencoder_inputs[entity_type], 1)

        # always holds the most-recent autoencoder reconstructions
        autoencoder_outputs = {}

        # always holds the most-recent bottleneck representations
        bottlenecks = torch.zeros(size=(num_entities, self._bottleneck_size), device=self._device)

        # zero-depth autoencoder
        depth = 0
        logging.debug("Running %d-depth autoencoder", depth)
        for entity_type, entity_index in entity_indices.items():
            entity_outputs, bns, losses = self._entity_autoencoders[entity_type][0](autoencoder_inputs[entity_type])
            autoencoder_outputs[entity_type] = entity_outputs
            bottlenecks[entity_index] = bns

        # n-depth autoencoders
        prev_bottlenecks = bottlenecks.clone()
        for depth in range(1, self._depth + 1):
            logging.debug("Running %d-depth autoencoder", depth)            
            for entity_type, entity_index in entity_indices.items():
                autoencoder_outputs[entity_type] = autoencoder_outputs[entity_type].narrow(1, 0, self._entity_autoencoders[entity_type][0].output_size)
                other_reps = []
                for rel_name in self._entity_type_relation_order[entity_type]:
                    summarize = self._summarizers[entity_type][rel_name]
                    relation_reps = torch.zeros(size=(len(entity_indices[entity_type]), self._bottleneck_size), device=self._device)
                    for i, index in enumerate(entity_indices[entity_type]):
                        if rel_name not in adjacencies:
                            continue
                        related_indices = index_space.masked_select(adjacencies[rel_name][index])
                        if len(related_indices) > 0:
                            logging.debug("%d has %s-related indices %s", index, rel_name, related_indices)
                            obns = torch.index_select(prev_bottlenecks, 0, related_indices)
                            relation_reps[i] = summarize(obns)
                    other_reps.append(relation_reps)
                sh = list(autoencoder_outputs[entity_type].shape)
                sh[1] = 0
                other_reps = torch.cat(other_reps, 1) if len(other_reps) > 0 else torch.zeros(size=tuple(sh), device=self._device)
                autoencoder_inputs[entity_type] = torch.cat([autoencoder_outputs[entity_type], other_reps], 1)
                if depth > len(self._entity_autoencoders[entity_type]) - 1:
                    entity_outputs, bns, losses = self._entity_autoencoders[entity_type][-1](autoencoder_inputs[entity_type])
                else:
                    entity_outputs, bns, losses = self._entity_autoencoders[entity_type][depth](autoencoder_inputs[entity_type])
                autoencoder_outputs[entity_type] = entity_outputs
                bottlenecks[entity_index] = bns

        logging.debug("Projecting autoencoder outputs")
        if self._depth > 0:
            resized_autoencoder_outputs = {k : self._projectors[k](v) for k, v in autoencoder_outputs.items()}
        else:
            resized_autoencoder_outputs = {k : v for k, v in autoencoder_outputs.items()}
            
        # reconstruct the entities by unfolding the last autoencoder output
        logging.debug("Reconstructing the input by applying decoders to the autoencoder output")
        reconstructions = {}
        for entity_type, entity_index in entity_indices.items():
            logging.debug("\tFor '%s' entities", entity_type)
            for field_name in self._entity_type_field_order.get(entity_type, []):
                logging.debug("\t\tFor field '%s'", field_name)
                out = self._field_decoders[field_name](resized_autoencoder_outputs[entity_type])
                if field_name not in reconstructions:
                    reconstructions[field_name] = torch.zeros(size=[num_entities] + list(out.shape)[1:], device=self._device)
                reconstructions[field_name][entity_mask] = out
        reconstructions[self._spec.id_field] = entities[self._spec.id_field]
        reconstructions[self.entity_type_field] = entities[self.entity_type_field]
        for rel in self.spec.relation_field_names:
            reconstructions[rel] = entities[rel]
        logging.debug("Returning reconstructions, bottlenecks, and autoencoder I/O pairs")
        return (reconstructions, bottlenecks, autoencoder_boundary_pairs)

    # Recursively initialize model weights
    def init_weights(m):
        if type(m) == torch.nn.Linear or type(m) == torch.nn.Conv1d:
            torch.nn.init.xavier_uniform_(m.weight)
            m.bias.data.fill_(0.01)




import math
import time
import calendar

class Field(object):
    def __init__(self, name, **args):
        self.name = name
        self.type_name = args["type"]
    def __str__(self):
        return "{1} field: {0}".format(self.name, self.type_name)
    def encode(self, v):
        return v
        #raise NotImplementedError()
    def decode(self, v):
        return v
        #raise NotImplementedError()        
    def observe_value(self, v):
        raise NotImplementedError()
    
class MetaField(Field):
    def __init__(self, name, **args):
        super(MetaField, self).__init__(name, **args)
    def encode(self, v):
        return v
    def decode(self, v):
        return v
    def observe_value(self, v):
        pass
        
class DataField(Field):        
    def __init__(self, name, **args):
        super(DataField, self).__init__(name, **args)
    
class EntityTypeField(MetaField):
    def __init__(self, name, **args):
        super(EntityTypeField, self).__init__(name, **args)

class RelationField(MetaField):
    def __init__(self, name, **args):
        super(RelationField, self).__init__(name, **args)
        self.target_entity_type = args["target_entity_type"]

class IdField(MetaField):
    def __init__(self, name, **args):
        super(IdField, self).__init__(name, **args)
        
class NumericField(DataField):    
    def __init__(self, name, **args):
        super(NumericField, self).__init__(name, **args)

class IntegerField(DataField):    
    def __init__(self, name, **args):
        super(IntegerField, self).__init__(name, **args)
        
class DateField(DataField):
    def __init__(self, name, **args):
        super(DateField, self).__init__(name, **args)
    def encode(self, v):
        t = time.strptime(v, "%d-%b-%Y")
        return calendar.timegm(t)
    def decode(self, v):
        date = time.gmtime(v)
        year = date.tm_year
        month = calendar.month_abbr[date.tm_mon]
        day = date.tm_mday
        return "{}-{}-{}".format(day, month, year)
    
class DistributionField(DataField):
    def __init__(self, name, **args):
        super(DistributionField, self).__init__(name, **args)

    def encode(self, v):
        total = sum(v.values())
        return [0.0 if c not in v else (v[c] / total) for c in self._categories]

    def decode(self, v):
        assert(len(v) == len(self._categories))
        retval = {}
        if all([x >= 0 for x in v]):
            total = sum([x for x in v])
            for k, p in zip(self._categories, v):
                if p > 0:
                    retval[k] = p / total
        elif all([x <= 0 for x in v]):            
            total = sum([math.exp(x) for x in v])
            for k, p in zip(self._categories, v):
                retval[k] = math.exp(p) / total            
        else:
            raise Exception("Got probabilities that were not all of the same sign!")
        return retval

class CategoricalField(DataField):

    def __init__(self, name, **args):
        super(CategoricalField, self).__init__(name, **args)
        self._lookup = {} #Missing() : 0}
        self._rlookup = {}

    def observe_value(self, v):
        i = self._lookup.setdefault(v, len(self._lookup))
        self._rlookup[i] = v
    
    def encode(self, v):
        return self.get(v, self.unknown_value)

    def decode(self, v):
        if v not in self._rlookup:
            raise Exception("Could not decode value '{0}' (type={2})".format(v, self._rlookup, type(v)))
        return self._rlookup[v]
    
    def __str__(self):
        return "{1} field: {0}[{2}]".format(self.name, self.type_name, len(self._lookup))

class SequentialField(DataField):

    def __init__(self, name, **args):
        super(SequentialField, self).__init__(name, **args)
        self._lookup = {}
        self._rlookup = {}
        #unique_sequences = set()
        #self[Missing] = Missing.value
        #self[Unknown] = Unknown.value
        #self._max_length = 0
        #for value in field_values:
        #    unique_sequences.add(value)
        #    self._max_length = max(self._max_length, len(value))
        #    for element in value:
        #        self[element] = self.get(element, len(self))
        #self._rlookup = {v : k for k, v in self.items()}
        #self._unique_sequence_count = len(unique_sequences)

    #def __str__(self):
    #    return "Sequential(unique_elems={}, unique_seqs={}, max_length={})".format(len(self),
    #                                                                               self._unique_sequence_count, 
    #                                                                               self._max_length)
    def observe_value(self, vs):
        for v in vs:
            i = self._lookup.setdefault(v, len(self._lookup))
            self._rlookup[i] = v
        
    def __str__(self):
        return "{1} field: {0}[{2}]".format(self.name, self.type_name, len(self._lookup))

    def encode(self, v):
        return [self.get(e, self[Unknown]) for e in v]

    def decode(self, v):
        try:
            return "".join([self._rlookup[e] for e in v if e not in [Missing.value, Unknown.value]])
        except:
            raise Exception("Could not decode values '{0}' (type={2})".format(v, self._rlookup, type(v[0])))


field_classes = {"numeric" : NumericField,
                 "categorical" : CategoricalField,
                 "boolean" : CategoricalField,                 
                 "sequential" : SequentialField,
                 "integer" : IntegerField,
                 "keyword" : CategoricalField,
                 "text" : SequentialField,
                 "relation" : RelationField,
                 "distribution" : DistributionField,
                 "date" : DateField,
                 "id" : IdField,
                 "entity_type" : EntityTypeField,
             }



import pickle
import re
import gzip
import sys
import argparse
import json
import random
import logging
import warnings
import numpy
import torch
from starcoder.dataset import Missing, Unknown


warnings.filterwarnings("ignore")


class Scheduler(torch.optim.lr_scheduler.ReduceLROnPlateau):
    def __init__(self, early_stop, *argv, **argdict):
        self.early_stop = early_stop
        self.num_bad_epochs_for_early_stop = 0
        super(Scheduler, self).__init__(*argv, **argdict)
        
    def step(self, metrics, epoch=None):
        is_reduce_rate = False
        is_early_stop = False
        is_new_best = False
        # convert `metrics` to float, in case it's a zero-dim Tensor
        current = float(metrics)
        if epoch is None:
            epoch = self.last_epoch + 1
        else:
            warnings.warn(EPOCH_DEPRECATION_WARNING, UserWarning)
        self.last_epoch = epoch

        if self.is_better(current, self.best):
            self.best = current
            is_new_best = True
            self.num_bad_epochs = 0
            self.num_bad_epochs_for_early_stop = 0
        else:
            self.num_bad_epochs += 1
            self.num_bad_epochs_for_early_stop += 1

        if self.in_cooldown:
            self.cooldown_counter -= 1
            self.num_bad_epochs = 0  # ignore any bad epochs in cooldown
            self.num_bad_epochs_for_early_stop = 0

        if self.num_bad_epochs > self.patience:
            self._reduce_lr(epoch)
            self.cooldown_counter = self.cooldown
            self.num_bad_epochs = 0
            is_reduce_rate = True

        if self.num_bad_epochs_for_early_stop > self.early_stop:
            is_early_stop = True

        self._last_lr = [group['lr'] for group in self.optimizer.param_groups]
        return (is_reduce_rate, is_early_stop, is_new_best)
    

def compute_losses(entities, reconstructions, spec, field_models):
    """
    Gather and return all the field losses in a dictionary.
    """
    losses = {}
    for field_name, entity_values in entities.items():
        #print(field_name)
        if field_name in reconstructions and field_name not in [spec.id_field, spec.entity_type_field] + list(spec.relation_field_names):
            mask = entity_values > 0
            field_type = type(spec.field_object(field_name))
            reconstruction_values = reconstructions[field_name]
            recon = torch.reshape(reconstruction_values, (reconstruction_values.shape[0], -1)).sum(1)
            selector = torch.masked_select(torch.arange(0, entity_values.shape[0]),
                                           ~torch.isnan(entity_values) & ~torch.isnan(recon))
            losses[field_type] = losses.get(field_type, {})
            masked_entity_values = torch.index_select(entity_values, 0, selector) #~torch.isnan(entity_values))

            masked_reconstruction_values = torch.index_select(reconstruction_values, 0, selector) #~torch.isnan(entity_values))
            #print(masked_entity_values.dtype, masked_reconstruction_values.dtype, field_models[field_type][2])
            masked_field_losses = field_models[field_type][2](masked_reconstruction_values, masked_entity_values)
            losses[field_type][field_name] = masked_field_losses
    return losses


def tensorize(vals, field_obj):
    if any([isinstance(v, list) for v in vals]):
        max_length = max([len(v) for v in vals if isinstance(v, list)])
        vals = [(v + ([Missing.value] * (max_length - len(v)))) if v != None else [Missing.value] * max_length for v in vals]
    elif any([isinstance(v, (str,)) for v in vals]):
        return numpy.array(vals)
    else:
        vals = [(field_obj.missing_value if v == None else v) for v in vals]
    return torch.tensor(vals, dtype=field_obj.encoded_type)


def stack_batch(components, spec):
    lengths = [len(x) for x, _ in components]
    entities = sum([x for x, _ in components], [])
    adjacencies = [x for _, x in components]
    full_adjacencies = {}
    start = 0
    for l, adjs in zip(lengths, adjacencies):
        for name, adj in adjs.items():
            full_adjacencies[name] = full_adjacencies.get(name, numpy.full((len(entities), len(entities)), False))
            full_adjacencies[name][start:start + l, start:start + l] = adj.todense()
        start += l
    field_names = spec.field_names #regular_field_names #set(sum([[k for k in e.keys()] for e in entities], []))
    #field_names.add(spec.id_field)
    full_entities = {k : [] for k in field_names}
    for entity in entities:
        for field_name in field_names:
            full_entities[field_name].append(entity.get(field_name, None))
    #full_entities = {k : tensorize(v, spec.field_object(k)) for k, v in full_entities.items()}
    full_entities = {k : numpy.array(v) if k in spec.relation_fields else tensorize(v, spec.field_object(k)) for k, v in full_entities.items()}
    ne = len(entities)
    for k, v in full_adjacencies.items():
        a, b = v.shape
        assert(ne == a and ne == b)
    return (full_entities, {k : torch.tensor(v) for k, v in full_adjacencies.items()})


def split_batch(entities, adjacencies, count):
    ix = list(range(len(entities)))
    random.shuffle(ix)
    first_ix = ix[0:count]
    second_ix = ix[count:]
    first_entities = [entities[i] for i in first_ix]
    second_entities = [entities[i] for i in second_ix]
    first_adjacencies = {}
    second_adjacencies = {}
    for rel_type, adj in adjacencies.items():
        adjacencies[rel_type] = adj[first_ix, :][:, first_ix]
        adjacencies[rel_type] = adj[second_ix, :][:, second_ix]
    return ((first_entities, first_adjacencies), (second_entities, second_adjacencies))



#
# This has some unpleasantly-complicated logic for different ways of 
# handling components that don't fit into a batch:
#
#   "strict" means "never create an oversized batch"
#   "subselect" means "it's OK to split components over multiple batches"
#
# If strict=True and subselect=False, components larger than the
# batch size will never be seen, partial or otherwise.
#
def batchify(data, batch_size, strict=True, subselect=False, mask_tests=[]):
    component_ids = range(data.num_components)
    _component_ids = [c for c in component_ids]
    random.shuffle(_component_ids)
    current_batch = []
    current_total = 0
    for component_id in _component_ids:
        full_entities, adjacencies = data.component(component_id)
        masked_entities = []
        for e in full_entities:
            masked_entities.append(e.copy())
            for k in list(e.keys()):
                for mask_test in mask_tests:
                    if mask_test(e, k):
                        del masked_entities[-1][k]
        full_encoded_entities = [data.encode(e) for e in full_entities]
        masked_encoded_entities = [data.encode(e) for e in masked_entities]
        encoded_entities = list(zip(full_encoded_entities, masked_encoded_entities))

        while len(encoded_entities) > 0:
            if len(encoded_entities) > batch_size and subselect == False:
                # component is larger than batch size and not subselecting
                if strict == False:
                    # if not strict, just yield it
                    yield(
                        (
                            stack_batch([([x for x, _ in encoded_entities], adjacencies)], data._spec),
                            stack_batch([([x for _, x in encoded_entities], adjacencies)], data._spec)
                        )
                    )
                entities = []                    
            elif current_total + len(encoded_entities) > batch_size:
                # component + current is larger than batch size
                if subselect == True:
                    (sub_encoded_entities, sub_adjacencies), (encoded_entities, adjacencies) = split_batch(encoded_entities, 
                                                                                                           adjacencies, 
                                                                                                           batch_size - current_total)
                    current_batch.append((sub_encoded_entities, sub_adjacencies))
                    yield(
                        (
                            stack_batch([([x for x, _ in c], a) for c, a in current_batch], data._spec),
                            stack_batch([([x for _, x in c], a) for c, a in current_batch], data._spec)
                        )
                    )
                    #yield(stack_batch(current_batch, data._spec))
                    current_batch, current_total = [], 0
                else:
                    yield(
                        (
                            stack_batch([([x for x, _ in c], a) for c, a in current_batch], data._spec),
                            stack_batch([([x for _, x in c], a) for c, a in current_batch], data._spec)
                        )
                    )
                    current_batch, current_total = [], 0
            else:
                # component + current not large enough yet
                current_batch.append((encoded_entities, adjacencies))
                current_total += len(encoded_entities)
                encoded_entities = []
    if len(current_batch) > 0:
        # final batch
        yield(
            (
                stack_batch([([x for x, _ in c], a) for c, a in current_batch], data._spec),
                stack_batch([([x for _, x in c], a) for c, a in current_batch], data._spec)
            )
        )


def run_over_components(model, field_models, optim, loss_policy, data, batch_size, gpu, train, subselect=False, strict=True, mask_tests=[]):
    old_mode = model.training
    model.train(train)
    loss_by_field = {}
    loss = 0.0
    for ((full_entities, full_adjacencies), (masked_entities, masked_adjacencies)) in batchify(data, batch_size, subselect=subselect, strict=strict, mask_tests=mask_tests):
        batch_loss_by_field = {}
        if gpu:
            entities = {k : v.cuda() for k, v in entities.items()}
            adjacencies = {k : v.cuda() for k, v in adjacencies.items()}
        optim.zero_grad()
        reconstructions, bottlenecks, ae_pairs = model(masked_entities, masked_adjacencies)
        for field_type, fields in compute_losses(full_entities, reconstructions, data._spec, field_models).items():
            for field_name, losses in fields.items():
                batch_loss_by_field[(field_name, field_type)] = losses
        batch_loss = loss_policy(batch_loss_by_field, ae_pairs)
        loss += batch_loss
        if train:
            batch_loss.backward()
            optim.step()
        for k, v in batch_loss_by_field.items():
            loss_by_field[k] = loss_by_field.get(k, [])
            loss_by_field[k].append(v.clone().detach())
    model.train(old_mode)
    return (loss, loss_by_field)


def run_epoch(model, field_models, optimizer, loss_policy, train_data, dev_data, batch_size, gpu, mask_tests=[]):
    model.train(True)
    train_loss, train_loss_by_field = run_over_components(model, 
                                                          field_models, 
                                                          optimizer, 
                                                          loss_policy,
                                                          train_data, 
                                                          batch_size, 
                                                          gpu, 
                                                          train=True,
                                                          mask_tests=mask_tests,
    )
    model.train(False)
    dev_loss, dev_loss_by_field = run_over_components(model, 
                                                      field_models, 
                                                      optimizer, 
                                                      loss_policy,
                                                      dev_data, 
                                                      batch_size, 
                                                      gpu, 
                                                      train=False,
                                                      mask_tests=mask_tests,
    )

    return (train_loss.clone().detach().cpu(), 
            {k : [v.clone().detach().cpu() for v in vv] for k, vv in train_loss_by_field.items()},
            dev_loss.clone().detach().cpu(),
            {k : [v.clone().detach().cpu() for v in vv] for k, vv in dev_loss_by_field.items()})



#!/usr/bin/env python

from distutils.core import setup

setup(name="starcoder",
      version="0.0.1",
      description="StarCoder is an unsupervised deep learning model tailored for exploratory empirical research in the humanities.",
      long_description="StarCoder is essentially a generator that combines autoencoder and graph-convolutional mechanisms with the open set of neural architectures to build end-to-end models of entity-relationship schemas.  By adopting intuitive JSON for all I/O, and using reconstruction loss as the objective, it allows researchers from other fields to apply it to their specific areas, while ML researchers can continue to augment it with new techniques.",
      long_description_content_type="text/x-rst",
      author="Thomas Lippincott",
      author_email="tom@cs.jhu.edu",
      url="https://github.com/TomLippincott/starcoder",
      maintainer="Thomas Lippincott",
      maintainer_email="tom@cs.jhu.edu",
      packages=["starcoder"],
      scripts=[],
      requires=["torch", "pytest", "sklearn", "scipy"],
     )



import pickle
import re
import sys
import argparse
import json
import gzip
import functools
import random
import logging
import numpy
import scipy.sparse
from sklearn.metrics import f1_score, accuracy_score
from scipy.sparse.csgraph import connected_components
import torch
import math
import uuid
from starcoder import fields
# import time
# import calendar

# class Field(object):
#     def __init__(self, name, **args):
#         self.name = name
#         self.type_name = args["type"]
#     def encode(self, v):
#         return v
#     def decode(self, v):
#         return v
#     def observe_value(self, v):
#         pass
#     def __str__(self):
#         return "{1} field: {0}".format(self.name, self.type_name)
    
# class MetaField(Field):
#     def __init__(self, name, **args):
#         super(MetaField, self).__init__(name, **args)
    
# class DataField(Field):        
#     def __init__(self, name, **args):
#         super(DataField, self).__init__(name, **args)
    
# class EntityTypeField(MetaField):
#     def __init__(self, name, **args):
#         super(EntityTypeField, self).__init__(name, **args)

# class RelationField(MetaField):
#     def __init__(self, name, **args):
#         super(RelationField, self).__init__(name, **args)
#         self.target_entity_type = args["target_entity_type"]

# class IdField(MetaField):
#     def __init__(self, name, **args):
#         super(IdField, self).__init__(name, **args)
        
# class NumericField(DataField):    
#     def __init__(self, name, **args):
#         super(NumericField, self).__init__(name, **args)

# class IntegerField(DataField):    
#     def __init__(self, name, **args):
#         super(IntegerField, self).__init__(name, **args)
        
# class DateField(DataField):
#     def __init__(self, name, **args):
#         super(DateField, self).__init__(name, **args)
#     def encode(self, v):
#         t = time.strptime(v, "%d-%b-%Y")
#         return calendar.timegm(t)
#     def decode(self, v):
#         date = time.gmtime(v)
#         year = date.tm_year
#         month = calendar.month_abbr[date.tm_mon]
#         day = date.tm_mday
#         return "{}-{}-{}".format(day, month, year)
    
# class DistributionField(DataField):
#     def __init__(self, name, **args):
#         super(DistributionField, self).__init__(name, **args)

#     def encode(self, v):
#         total = sum(v.values())
#         return [0.0 if c not in v else (v[c] / total) for c in self._categories]

#     def decode(self, v):
#         assert(len(v) == len(self._categories))
#         retval = {}
#         if all([x >= 0 for x in v]):
#             total = sum([x for x in v])
#             for k, p in zip(self._categories, v):
#                 if p > 0:
#                     retval[k] = p / total
#         elif all([x <= 0 for x in v]):            
#             total = sum([math.exp(x) for x in v])
#             for k, p in zip(self._categories, v):
#                 retval[k] = math.exp(p) / total            
#         else:
#             raise Exception("Got probabilities that were not all of the same sign!")
#         return retval


# # class IntegerField(DataField):

# #     name = "numeric"

# #     def __init__(self, field_values):
# #         field_values = [float(x) for x in field_values]
# #         self._minimum = min(field_values)
# #         self._maximum = max(field_values)
# #         self._trivial = self._minimum == self._maximum

# #     def __str__(self):
# #         return "Integer(min/max={}/{})".format(self._minimum, self._maximum)

# #     def encode(self, v):
# #         return (0.0 if self._trivial else (v - self._minimum) / (self._maximum - self._minimum))

# #     def decode(self, v):
# #         return round(self._minimum if self._trivial else (v * (self._maximum - self._minimum) + self._minimum))


# class Missing(object):
#     pass

# class Unknown(object):
#     pass

    

# class CategoricalField(DataField):

#     #name = "categorical"
#     #missing_value = 0
#     #unknown_value = 1
#     #encoded_type = int
#     #na_value = 0
    
#     def __init__(self, name, **args):
#         super(CategoricalField, self).__init__(name, **args)
#         self._lookup = {} #Missing() : 0}
#         self._rlookup = {}
#         #self[Missing()] = self.missing_value #Missing.value
#         #self[Unknown()] = self.unknown_value
#         #for value in field_values:
#         #    self[value] = self.get(value, len(self))
#         #self._rlookup = {v : k for k, v in self.items()}

#     #def __str__(self):
#     #    return "Categorical(possible_values={})".format(len(self))

#     def observe_value(self, v):
#         i = self._lookup.setdefault(v, len(self._lookup))
#         self._rlookup[i] = v

    
#     def encode(self, v):
#         return self.get(v, self.unknown_value)

#     def decode(self, v):
#         if v not in self._rlookup:
#             raise Exception("Could not decode value '{0}' (type={2})".format(v, self._rlookup, type(v)))
#         return self._rlookup[v]
    
#     def __str__(self):
#         return "{1} field: {0}[{2}]".format(self.name, self.type_name, len(self._lookup))

# class SequentialField(DataField):

#     #name = "sequential"

#     def __init__(self, name, **args):
#         super(SequentialField, self).__init__(name, **args)
#         self._lookup = {}
#         self._rlookup = {}
#         #unique_sequences = set()
#         #self[Missing] = Missing.value
#         #self[Unknown] = Unknown.value
#         #self._max_length = 0
#         #for value in field_values:
#         #    unique_sequences.add(value)
#         #    self._max_length = max(self._max_length, len(value))
#         #    for element in value:
#         #        self[element] = self.get(element, len(self))
#         #self._rlookup = {v : k for k, v in self.items()}
#         #self._unique_sequence_count = len(unique_sequences)

#     #def __str__(self):
#     #    return "Sequential(unique_elems={}, unique_seqs={}, max_length={})".format(len(self),
#     #                                                                               self._unique_sequence_count, 
#     #                                                                               self._max_length)
#     def observe_value(self, vs):
#         for v in vs:
#             i = self._lookup.setdefault(v, len(self._lookup))
#             self._rlookup[i] = v
        
#     def __str__(self):
#         return "{1} field: {0}[{2}]".format(self.name, self.type_name, len(self._lookup))

#     def encode(self, v):
#         return [self.get(e, self[Unknown]) for e in v]

#     def decode(self, v):
#         try:
#             return "".join([self._rlookup[e] for e in v if e not in [Missing.value, Unknown.value]])
#         except:
#             raise Exception("Could not decode values '{0}' (type={2})".format(v, self._rlookup, type(v[0])))


# field_classes = {"numeric" : NumericField,
#                  "categorical" : CategoricalField,
#                  "boolean" : CategoricalField,                 
#                  "sequential" : SequentialField,
#                  #"integer" : IntegerField,
#                  "keyword" : CategoricalField,
#                  "text" : SequentialField,
#                  "relation" : RelationField,
#                  "distribution" : DistributionField,
#                  "date" : DateField,
#                  "id" : IdField,
#                  "entity_type" : EntityTypeField,
#              }


class Datum(dict):
    def __init__(self, schema, obj):
        self._meta_fields = {}
        self._data_fields = {}
        self._relation_fields = {}
        self._unknown_fields = {}
        for k, v in obj.items():
            pass

class EntityType(object):
    def __init__(self, name, data_fields, relation_fields):
        self.name = name
        self.data_fields = data_fields
        self.relation_fields = relation_fields
    def __str__(self):
        return "{}: data fields={}, relation fields={}".format(self.name, self.data_fields, self.relation_fields)

class Schema(object):
    def __init__(self, spec):
        self.id_field_name = spec["meta"]["id_field"]
        self.entity_type_field_name = spec["meta"]["entity_type_field"]
        self._field_objects = {self.id_field_name : fields.field_classes["id"](self.id_field_name),
                               self.entity_type_field_name : fields.field_classes["entity_type"](self.entity_type_field_name)
                               }
        self._entity_relation_field_names = {}
        self._entity_data_field_names = {}
        self._relation_field_names = set()
        self._data_field_names = set()        
        for field_name, field_spec in spec["data_fields"].items():
            field_type = field_spec["type"]
            self._data_field_names.add(field_name)
            self._field_objects[field_name] = field.field_classes[field_type](field_name, **field_spec)

        for entity_type, fields in spec["entity_types"].items():
            self._entity_data_field_names[entity_type] = set(fields)
            self._entity_relation_field_names[entity_type] = set()
            
        for field_name, field_spec in spec["relation_fields"].items():
            self._relation_field_names.add(field_name)
            self._entity_relation_field_names[entity_type].add(field_name)
            self._field_objects[field_name] = field_classes["relation"](field_name, type="relation", **field_spec)            

        self._entity_types = {}
        for entity_type, data_fields in self._entity_data_field_names.items():
            self._entity_types[entity_type] = EntityType(entity_type,
                                                         data_fields,
                                                         self._entity_relation_field_names.get(entity_type, []))

    @property
    def entity_type_field(self):
        return self._field_objects[self.entity_type_field_name]

    @property
    def id_field(self):
        return self._field_objects[self.id_field_name]

    @property
    def entity_types(self):
        return self._entity_types
            
    def __str__(self):
        return """
Schema(
    Fields:
        {}
        {} 
        {}
        {}
    entity types:
        {}
)
""".format(self._field_objects[self.id_field_name],
           self._field_objects[self.entity_type_field_name],
           "\n        ".join([str(self._field_objects[f]) for f in self._data_field_names]),
           "\n        ".join([str(self._field_objects[f]) for f in self._relation_field_names]),
           "\n        ".join([str(et) for et in self._entity_types.values()])
)

    def observe_entity(self, entity):
        for k, v in entity.items():
            #print(k)
            if k in self._field_objects:
                self._field_objects[k].observe_value(v)

                
    
    # def entity_relations(self, entity_type):
    #     return set([x for x in self.entity_fields(entity_type) if isinstance(self.field_object(x), RelationField)])

    # @property
    # def regular_field_names(self):
    #     return set([f for f, o in self._field_name_to_object.items() if not isinstance(o, (IdField, RelationField))])

    # @property
    # def entity_types(self):
    #     return set(self._entity_type_to_field_names.keys())

    # def entity_fields(self, entity_name):
    #     return self._entity_type_to_field_names.get(entity_name, set())

    # def field_object(self, field_name):
    #     return self._field_name_to_object[field_name]

    # def exists(self, field_name):
    #     return (field_name in self._field_name_to_object)

    # @property
    # def field_names(self):
    #     return set([f for f in self._field_name_to_object.keys()])

    # @property
    # def data_field_names(self):
    #     return set([f for f in self._field_name_to_object.keys() if f not in self._relation_fields and f not in [self._id_field, self._entity_type_field]])

    # @property
    # def data_fields(self):
    #     return set([f for f in self._field_name_to_object.keys() if f not in self._relation_fields and f not in [self._id_field, self._entity_type_field]])
           
    # @property
    # def entity_type_field(self):
    #     return self._entity_type_field

    # @property
    # def id_field_name(self):
    #     return self._id_field

    # @property
    # def id_field(self):
    #     return self._id_field

    # @property
    # def relation_field_names(self):
    #     return self._relation_fields

    # @property
    # def relation_fields(self):
    #     return self._relation_fields
           
    # @property
    # def has_id_field(self):
    #     return (self._id_field != None)

    # @property
    # def num_fields(self):
    #     return len(self._field_name_to_object)

    # @property
    # def num_entities(self):
    #     return len(self._entity_type_to_field_names)

    # @property
    # def num_relations(self):
    #     return sum([len(x) for x in self._entity_type_to_relation_types.values()])

    # def encode(self, datum):
    #     retval = {}
    #     for k, v in datum.items():
    #         if k in self.entity_relations(datum[self.entity_type_field]):
    #             retval[k] = self._field_name_to_object[self.id_field_name].encode(v)
    #         elif k not in self._field_name_to_object:
    #             retval[k] = v
    #         else:
    #             retval[k] = self._field_name_to_object[k].encode(v)
    #     return retval

    # def decode(self, datum):
    #     retval = {}
    #     entity_type = self._field_name_to_object[self.entity_type_field].decode(datum[self.entity_type_field])
    #     for k, v in datum.items():
    #         if k in self.entity_relations(entity_type):
    #             retval[k] = self._field_name_to_object[self.id_field_name].decode(v)
    #         elif k not in self._field_name_to_object:
    #             retval[k] = v
    #         else:
    #             try:
    #                 retval[k] = self._field_name_to_object[k].decode(v)
    #             except:
    #                 raise Exception("Could not decode {} value '{}' (spec={})".format(k, v, self._field_name_to_object[k]))
    #     return retval

    # def decode_batch(self, batch):
    #     retvals = []
    #     for i in range(len(batch[self.entity_type_field])):
    #         cur = {}
    #         for field_name, values in batch.items():
    #             if values[i] != None:
    #                 if isinstance(values[i], str):
    #                     cur[field_name] = values[i]
    #                 else:
    #                     cur[field_name] = values[i].tolist()
    #         dec = self.decode(cur)
    #         if dec[self.id_field] != None:
    #             retvals.append({k : v for k, v in dec.items() if k in list(self.entity_fields(dec[self.entity_type_field])) + [None]})
    #     return retvals

    # def encode_batch(self, batch):
    #     retval = {}
    #     fields = set()
    #     for entity in batch:
    #         for k in entity.keys():
    #             retval[k] = retval.get(k, torch.empty(size=(len(batch),)))
    #     for i, entity in enumerate(batch):
    #         for k, v in self.encode(entity).items():
    #             retval[k][i] = v
    #     return retval
    

class Dataset(object):
    """
    The Dataset class is neede mainly for operations that depend on
    graph structure, particularly those that require connected components.
    """
    def __init__(self, schema, entities):
        self.schema = schema
        print(self.schema)
        self._entities = []
        self._entity_fields = {}
        self._id_to_index = {}
        for idx, entity in enumerate(entities):
            #entity_type = entity[self._spec.entity_type_field]
            #entity_id = entity[self._spec.id_field]
            #assert (entity_id not in self._id_to_index), "The id field ('{}') must be unique".format(self._spec.id_field)
            self._id_to_index[entity[self.id_field.name]] = idx
            #self._entity_fields[entity_type] = self._spec.entity_fields(entity_type) #
            self._entities.append(entity)
        self._edges = {}
        for entity in self._entities:
            entity_type = entity[self.entity_type_field.name]
            entity_id = entity[self.id_field.name]
            source_index = self._id_to_index[entity_id]
            for relation_type in self.entity_types[entity_type].relation_fields:
                target_ids = entity.get(relation_type, [])
                for target in target_ids if isinstance(target_ids, list) else [target_ids]:
                    if target not in self._id_to_index:
                        logging.warning("Could not find target %s for entity %s relation %s", target, entity_id, relation_type)
                        continue
                    target_index = self._id_to_index[target]
                    self._edges[relation_type] = self._edges.get(relation_type, {})
                    self._edges[relation_type][source_index] = self._edges[relation_type].get(source_index, [])
                    self._edges[relation_type][source_index].append(target_index)
        self._update_components()

    @property
    def entity_types(self):
        return self.schema.entity_types
        
    @property
    def entity_type_field(self):
        return self.schema.entity_type_field

    @property
    def id_field(self):
        return self.schema.id_field        
        
    def subselect(self, indices):
        return Dataset(self._schema, [self._entities[i] for i in indices])

    def subselect_components(self, indices):
        print(self.schema)
        return Dataset(self.schema, [self._entities[j] for j in sum([self._components[i][0] for i in indices], [])])
    
    def encode(self, item):
        return self._spec.encode(item)

    def decode(self, item):
        return self._spec.decode(item)

    def _update_components(self):
        # create union adjacency matrix
        rows, cols, vals = [], [], []
        for _, rs in self._edges.items():
            for r, cs in rs.items():
                for c in cs:
                    rows.append(r)
                    cols.append(c)
                    vals.append(True)
        adjacency = scipy.sparse.csr_matrix((vals, (rows, cols)), 
                                            shape=(len(self), len(self)), 
                                            dtype=numpy.bool)

        # create list of connected components
        num, ids = connected_components(adjacency)
        components = {}    
        for i, c in enumerate(ids):
            components[c] = components.get(c, [])
            components[c].append(i)
            
        largest_component_size = 0 if len(components) == 0 else max([len(x) for x in components.values()])
        if len(components) == 0:
            raise Exception("The data is empty: this probably isn't what you want.  Perhaps add more instances, or adjust the train/dev/test split proportions?")
        logging.info("Found %d connected components with maximum size %d", len(components), largest_component_size)
        
        self._components = []
        for c in components.values():
            #component_adjacencies = {} #{k : numpy.full((len(c), len(c)), False) for k in self._edges.keys()}
            ca_rows, ca_cols = {}, {}
            g2l = {k : v for v, k in enumerate(c)}
            for gsi in c:
                lsi = g2l[gsi]
                for rel_type, rows in self._edges.items():
                    ca_rows[rel_type] = ca_rows.get(rel_type, [])
                    ca_cols[rel_type] = ca_cols.get(rel_type, [])
                    for gti in rows.get(gsi, []):                        
                        lti = g2l[gti]
                        ca_rows[rel_type].append(lsi)
                        ca_cols[rel_type].append(lti)
                        #component_adjacencies[rel_type][lsi, lti] = True
            component_adjacencies = {rel_type : scipy.sparse.csr_matrix(([True for _ in ca_rows[rel_type]], (ca_rows[rel_type], ca_cols[rel_type])), 
                                                                 shape=(len(c), len(c)), dtype=numpy.bool) for rel_type in self._edges.keys()}
            self._components.append((c, component_adjacencies)) # = [c for c in components.values()]

        
    def __getitem__(self, index):
        return self._entities[index]
    
    def __len__(self):
        return len(self._entities)

    def component(self, i):
        entity_indices, adjacencies = self._components[i]
        entities = [self[i] for i in entity_indices]
        return (entities, adjacencies)

    #@property
    #def spec(self):
    #    return self._spec
    
    #@property
    #def entity_types(self):
    #    return self.spec.entity_types
    
    #@property
    #def field_names(self):
    #    return [x for x in self._spec.keys() if isinstance(x, str)]

    @property
    def num_components(self):
        return len(self._components)

    def __str__(self):
        return "Dataset({} entities, {} components with max size {})".format(len(self),
                                                                             len(self._components),
                                                                             0 if len(self._components) == 0 else max([len(c) for c, _ in self._components]))


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("--input", dest="input", help="Input file")
    parser.add_argument("--spec", dest="spec", help="Input file")
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    with gzip.open(args.input, "rb") as ifd:
        data = pickle.load(ifd)
        
    with gzip.open(args.spec, "rb") as ifd:
        spec = pickle.load(ifd)


